import os
from typing import List, Tuple, Any, Union, Optional, Dict
import mlflow
import mlflow.sklearn
from  mlflow.tracking import MlflowClient
from kaml.client import *
import pandas as pd
import numpy as np
import seaborn as sn
from sklearn.metrics import accuracy_score,roc_auc_score,log_loss,recall_score,f1_score,precision_score
import warnings
from IPython.utils import io
import logging
import traceback

#mlflow.sklearn.autolog(log_models=True)
mlflow.set_tracking_uri(os.environ['MLFLOW_BASE_URL'])
client = mlflow.tracking.MlflowClient(os.environ['MLFLOW_BASE_URL'])
logging.basicConfig(level=logging.WARN)
logger = logging.getLogger(__name__)

warnings.filterwarnings("ignore")  
         
# 'rbfsvm' - SVM - Radial Kernel               
# 'gpc' - Gaussian Process Classifier                  
# 'mlp' - MLP Classifier                   
# 'qda' - Quadratic Discriminant Analysis                  
# 'lda' - Linear Discriminant Analysis                  
# 'et' - Extra Trees Classifier   

def input_validation(X_train:pd.DataFrame,
                     X_test:pd.DataFrame,
                     y_train:pd.Series,
                     y_test:pd.Series,
                     experiment_name:str
                         )->None:
    if not isinstance(X_train, pd.DataFrame):
        raise TypeError(f"X_train passed must be of type pandas.DataFrame")
    if X_train.shape[0] == 0:
        raise ValueError(f"X_train passed must be a positive dataframe")
    if not isinstance(X_test, pd.DataFrame):
        raise TypeError(f"X_test passed must be of type pandas.DataFrame")
    if X_test.shape[0] == 0:
        raise ValueError(f"X_test passed must be a positive dataframe")
    if not (isinstance(y_train,pd.Series) or isinstance(y_train,pd.DataFrame)or isinstance(y_train,np.ndarray)):
        raise TypeError(f"y_train passed must be of type pandas.Series,pandas.DataFrame,numpy.ndarray") 
    if y_train.shape[0] == 0:
        raise ValueError(f"y_train passed must be a positive Series")
    if not (isinstance(y_test,pd.Series) or isinstance(y_test,pd.DataFrame)or isinstance(y_test,np.ndarray)):
        raise TypeError(f"y_test passed must be of type pandas.Series,pandas.DataFrame,numpy.ndarray")
    if y_test.shape[0] == 0:
        raise ValueError(f"y_test passed must be a positive Series")
    if X_train.shape[0]!=len(y_train):
        raise ValueError(f"Length of Train set and its respective Target variable should be of equal length")
    if X_test.shape[0]!=len(y_test):
        raise ValueError(f"Length of Test set and its respective Target variable should be of equal length")
    if not isinstance(experiment_name, str):
        raise ValueError(f"experiment_name must be string")
    if experiment_name=="":
        raise ValueError(f"experiment_name cannot be null or empty string")
        
class Classifier(MLClient):
    
    def __init__(self, 
                 X_train:pd.DataFrame,
                 X_test:pd.DataFrame,
                 y_train:pd.Series,
                 y_test:pd.Series,
                 experiment_name:str):
        '''
        Parameters
        ----------
        X_train:pandas.DataFrame
            Shape (n_samples, n_features), where n_samples is the number of samples and 
            n_features is the number of features for train dataset.

        X_test:pandas.DataFrame
            Shape (n_samples, n_features), where n_samples is the number of samples and 
            n_features is the number of features for test dataset

        y_train:pandas.Series
           Shape (n_samples,1),where n_samples is the number of samples for train dataset

        y_test:pandas.Series
           Shape (n_samples,1),where n_samples is the number of samples for test dataset

        experiment_name:str
           Name of the experiment for logging.
           
        '''
        
        logger.info(f"Validation Start")
        input_validation(X_train,X_test, y_train,y_test,experiment_name)
        logger.info(f"Validation End")
        
        super().__init__(experiment_name)
        
        self.X_train = X_train
        self.X_test = X_test
        self.y_train = y_train
        self.y_test = y_test
        
        logger.info(f"Classification successfully instantiated")

    def __str__(self):
        return "Classfiaction models for experiment: Experiment Name is %s, Experiment ID is %s" % (self.name, self.id)
    
    def eval_metrics(self, actual, pred):

        '''
        Calculates various evaluation parameters like accuracy_score,roc_auc_score,log_loss,recall_score,f1_score,precision_score for classification  problem

        Parameters:
        actual (pd.Dataframe) : actual y value
        pred (pd.Dataframe) : predicted y value

        Return:
        acc_score : Accuracy score
        auc_roc : roc auc Score
        log_los : log_loss
        recall : recall
        f1_scr : F1 Score
        precision_scr : precision Score 
        '''
        logger.info(f"Calculating evaluation paramaters")
        '''
        try:
            logger.info(f"classification Report")
            logger.info(classification_report(actual, pred))
        except:
            logger.error(f"classification_report calcuation raised an exception:")
            logger.error(traceback.format_exc())
        try:
            logger.info(f"Confusion Matrix")
            confusion_matrix = pd.crosstab(actual, pred, rownames=['Actual'], colnames=['Predicted'])
            logger.info(sn.heatmap(confusion_matrix, annot=True))
        except:
            logger.error(f"confusion matrix calcuation raised an exception:")
            logger.error(traceback.format_exc())
        '''
        try:
            acc_score = accuracy_score(actual, pred)
        except:
            logger.error(f"accuracy_score calcuation raised an exception:")
            logger.error(traceback.format_exc())
        try:
            auc_roc = roc_auc_score(actual, pred)
        except:
            logger.error(f"roc_auc_score calcuation raised an exception:")
            logger.error(traceback.format_exc())
        try:
            log_los = log_loss(actual, pred)
        except:
            logger.error(f"log_loss calcuation raised an exception:")
            logger.error(traceback.format_exc())
        try:
            recall = recall_score(actual, pred)
        except:
            logger.error(f"recall_score calcuation raised an exception:")
            logger.error(traceback.format_exc())
        try:
            f1_scr = f1_score(actual, pred)
        except:
            logger.error(f"f1_score calcuation raised an exception:")
            logger.error(traceback.format_exc())
        try:
            precision_scr = precision_score(actual, pred)
        except:
            logger.error(f"precision_score calcuation raised an exception:")
            logger.error(traceback.format_exc())

        return acc_score,auc_roc,log_los,recall,f1_scr,precision_scr

    def model_fit_log(self, model, run_name:str)->None:
        '''
        Performs the model fitting and model prediction task. It calls the evaluation metrics function and logs the required metrics for the model.

        Parameters:
        model (Object) : classification model object
        '''
        logger.info(f"Training the model {run_name}")
        try:
            model.fit(self.X_train,self.y_train)
        except:
            logger.error(f"Fitting the model {model} raised an exception:")
            logger.error(traceback.format_exc())

        logger.info(f"Performing the prediction")

        try:
            y_pred=model.predict(self.X_test)
        except:
            logger.error(f" Prediction process raised an exception:")
            logger.error(traceback.format_exc())

        (acc_score,auc_roc,log_los,recall,f1_scr,precision_scr) = self.eval_metrics(self.y_test, y_pred)

        metrics = {
            "accuracy_score" : acc_score,
            "roc_auc_score" : auc_roc,
            "log_loss" : log_los,
            "recall" : recall,
            "f1_score" : f1_scr,
            "precision_score" : precision_scr
            }

        logger.info(f"logged params: {metrics.keys()}")
        try:
            mlflow.log_metrics(metrics)
        except:
            logger.warning("Couldn't perform log_metrics. Exception:")
            logger.warning(traceback.format_exc())

        logger.info(f"log model {run_name}")
        try:
            mlflow.sklearn.log_model(model, run_name)
        except:
            logger.warning("Couldn't log model. Exception:")
            logger.warning(traceback.format_exc())


    def LogisticRegression(self,**kwargs)->None:
        ''' This module performs experiments for Logistic Regression model

        Parameters
        ----------
        **kwargs: 
           List of parameters that will be passed to define the model.
           Refer: https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LogisticRegression.html
           for list of parameters that can be passed to model

         Returns
         -------
         None
         ''' 
        logger.info(f"-----------Logistic Regression--------------")
        
        run_name = self.name+ "_"+self.id + "_logistic_regression"

        with mlflow.start_run(run_name=run_name):
            from sklearn.linear_model import LogisticRegression
            try:
                model = LogisticRegression(**kwargs)
                self.model_fit_log(model,run_name)
            except:
                logger.warning("Couldn't perform runs on model. Exception:")
                logger.warning(traceback.format_exc())

    def RandomForestClassifier(self,**kwargs
                         )->None:
        ''' This module performs experiments for Random Forest Classifier model

        Parameters
        ----------
        **kwargs: 
           List of parameters that will be passed to define the model.
           Refer: https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.RandomForestClassifier.html
           for list of parameters that can be passed to model

         Returns
         -------
                 None
        '''
        logger.info(f"-----------Random Forest Classifier--------------")
        run_name = self.name+ "_"+self.id + "_random_forest_classifier"

        with mlflow.start_run(run_name=run_name):
            from sklearn.ensemble import RandomForestClassifier
            try:
                model =RandomForestClassifier(**kwargs)
                self.model_fit_log(model,run_name)
            except:
                logger.warning("Couldn't perform runs on model. Exception:")
                logger.warning(traceback.format_exc())


    def AdaBoostClassifier(self, **kwargs
                         )->None:
        ''' This module performs experiments for Ada Boost Classifier model

        Parameters
        ----------
        **kwargs: 
           List of parameters that will be passed to define the model.
           Refer: https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.AdaBoostClassifier.html
           for list of parameters that can be passed to model

         Returns
         -------
                 None
        '''
        logger.info(f"-----------Ada Boost Classifier--------------")
        
        run_name = self.name+ "_"+self.id +  "_ada_boost_classifier"

        with mlflow.start_run(run_name=run_name):
            from sklearn.ensemble import AdaBoostClassifier
            try:
                model =AdaBoostClassifier(**kwargs)
                self.model_fit_log(model,run_name)
            except:
                logger.warning("Couldn't perform runs on model. Exception:")
                logger.warning(traceback.format_exc())

    def GradientBoostingClassifier(self,**kwargs
                         )->None:
        ''' This module performs experiments for Gradient Boosting Classifier model

        Parameters
        ----------
        **kwargs: 
           List of parameters that will be passed to define the model.
           Refer: https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.GradientBoostingClassifier.html
           for list of parameters that can be passed to model

         Returns
         -------
           None
        '''
        logger.info(f"-----------Gradient Boosting Classifier--------------")
        
        run_name = self.name+ "_"+self.id +  "_gradient_boosting_classifier"

        with mlflow.start_run(run_name=run_name):
            from sklearn.ensemble import GradientBoostingClassifier
            try:
                model =GradientBoostingClassifier(**kwargs)
                self.model_fit_log(model,run_name)
            except:
                logger.warning("Couldn't perform runs on model. Exception:")
                logger.warning(traceback.format_exc())

    def CatBoostClassifier(self, **kwargs
                         )->None:
        ''' This module performs experiments for Cat Boost Classifier model

        Parameters
        ----------
        **kwargs: 
           List of parameters that will be passed to define the model.
           Refer: https://catboost.ai/en/docs/concepts/python-usages-examples
           for list of parameters that can be passed to model

         Returns
         -------
                 None
        '''
        logger.info(f"-----------Cat Boost Classifier--------------")
        
        run_name = self.name+ "_"+self.id +  "_cat_boost_classifier"

        with mlflow.start_run(run_name=run_name):
            from catboost import CatBoostClassifier
            try:
                model =CatBoostClassifier(**kwargs)
                self.model_fit_log(model,run_name)
            except:
                logger.warning("Couldn't perform runs on model. Exception:")
                logger.warning(traceback.format_exc())


    def LGBMClassifier(self,**kwargs)->None:
        ''' This module performs experiments for LightGBM Classifier model

        Parameters
        ----------
        **kwargs: 
           List of parameters that will be passed to define the model.
           Refer: https://lightgbm.readthedocs.io/en/latest/pythonapi/lightgbm.LGBMClassifier.html
           for list of parameters that can be passed to model

         Returns
         -------
                 None
        '''
        logger.info(f"-----------Light GBM Classifier--------------")
        
        run_name = self.name+ "_"+self.id+"_"+"lgbm_classifier"

        with mlflow.start_run(run_name=run_name):
            from lightgbm import LGBMClassifier
            try:
                model =LGBMClassifier(**kwargs)
                self.model_fit_log(model,run_name)
            except:
                logger.warning("Couldn't perform runs on model. Exception:")
                logger.warning(traceback.format_exc())


    def XGBClassifier(self, **kwargs)->None:
        ''' This module performs experiments for XGBoost Classifier model

        Parameters
        ----------
        **kwargs: 
           List of parameters that will be passed to define the model.
           Refer: https://xgboost.readthedocs.io/en/latest/python/python_api.html#xgboost.XGBClassifier
           for list of parameters that can be passed to model

         Returns
         -------
                 None
        '''
        logger.info(f"-----------XG Boost Classifier--------------")
    
        run_name = self.name+ "_"+self.id+"_"+"xgb_classifier"

        with mlflow.start_run(run_name=run_name):
            from xgboost import XGBClassifier
            try:
                model =XGBClassifier(**kwargs)
                self.model_fit_log(model,run_name)
            except:
                logger.warning("Couldn't perform runs on model. Exception:")
                logger.warning(traceback.format_exc())

    def DecisionTreeClassifier(self, **kwargs)->None:
        ''' This module performs experiments for Decision Tree Classifier model

        Parameters
        ----------
        **kwargs: 
           List of parameters that will be passed to define the model.
           Refer: https://scikit-learn.org/stable/modules/generated/sklearn.tree.DecisionTreeClassifier.html
           for list of parameters that can be passed to model

         Returns
         -------
                 None
        '''
        logger.info(f"-----------Decision Tree Classifier--------------")

        run_name = self.name+ "_"+self.id+"_"+"decision_tree_classifier"

        with mlflow.start_run(run_name=run_name):
            from sklearn.tree import DecisionTreeClassifier
            try:
                model =DecisionTreeClassifier(**kwargs)
                self.model_fit_log(model,run_name)
            except:
                logger.warning("Couldn't perform runs on model. Exception:")
                logger.warning(traceback.format_exc())


    def SupportVectorClassifier(self, **kwargs)->None:
        ''' This module performs experiments for Support Vector Classifier model

        Parameters
        ----------
        **kwargs: 
           List of parameters that will be passed to define the model.
           Refer: https://scikit-learn.org/stable/modules/svm.html
           for list of parameters that can be passed to model

         Returns
         -------
                 None
        '''
        logger.info(f"-----------Support Vector Machine Classifier--------------")

        run_name = self.name+ "_"+self.id+"_"+"svm_classifier"

        with mlflow.start_run(run_name=run_name):
            from sklearn.svm import SVC
            try:
                model = SVC(**kwargs)
                self.model_fit_log(model,run_name)
            except:
                logger.warning("Couldn't perform runs on model. Exception:")
                logger.warning(traceback.format_exc())

    def RidgeClassifier(self, **kwargs)->None:
        ''' This module performs experiments for Ridge Classifier model

        Parameters
        ----------
        **kwargs: 
           List of parameters that will be passed to define the model.
           Refer: https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.RidgeClassifier.html
           for list of parameters that can be passed to model

         Returns
         -------
                 None
        '''
        logger.info(f"-----------Ridge Classifier--------------")
    
        run_name = self.name+ "_"+self.id+"_"+"ridge_classifier"

        with mlflow.start_run(run_name=run_name):
            from sklearn.linear_model import RidgeClassifier
            try:
                model = RidgeClassifier(**kwargs)
                self.model_fit_log(model,run_name)
            except:
                logger.warning("Couldn't perform runs on model. Exception:")
                logger.warning(traceback.format_exc())

    def KNeighborsClassifier(self, **kwargs)->None:
        ''' This module performs experiments for K Neighbors Classifier model

        Parameters
        ----------
        **kwargs: 
           List of parameters that will be passed to define the model.
           Refer: https://scikit-learn.org/stable/modules/generated/sklearn.neighbors.KNeighborsClassifier.html
           for list of parameters that can be passed to model

         Returns
         -------
                 None
        '''
        logger.info(f"-----------K Neighbors Classifier--------------")
        
        run_name = self.name+ "_"+self.id+"_"+"k_neighbors_classifier"

        with mlflow.start_run(run_name=run_name):
            from sklearn.neighbors import KNeighborsClassifier
            try:
                model = KNeighborsClassifier(**kwargs)
                self.model_fit_log(model,run_name)
            except:
                logger.warning("Couldn't perform runs on model. Exception:")
                logger.warning(traceback.format_exc())

    def GaussianNB(self,**kwargs)->None:
        ''' This module performs experiments for Gaussian Naive Bayes Classifier model

        Parameters
        ----------
        **kwargs: 
           List of parameters that will be passed to define the model.
           Refer: https://scikit-learn.org/stable/modules/generated/sklearn.naive_bayes.GaussianNB.html
           for list of parameters that can be passed to model

         Returns
         -------
                 None
        '''
        logger.info(f"-----------Gaussian Naive Bayes Classifier--------------")
        
        run_name = self.name+ "_"+self.id+"_"+"gaussian_NB_classifier"

        with mlflow.start_run(run_name=run_name):
            from sklearn.naive_bayes import GaussianNB
            try:
                model = GaussianNB(**kwargs)
                self.model_fit_log(model,run_name)
            except:
                logger.warning("Couldn't perform runs on model. Exception:")
                logger.warning(traceback.format_exc())
