
## This regression file acts as a central repository for various regression models that is being packaged under kaml module

import os
from typing import List
import mlflow
import mlflow.sklearn
from  mlflow.tracking import MlflowClient
from kaml.client import MLClient
import pandas as pd
import numpy as np
import sklearn
from sklearn.metrics import mean_squared_error, mean_absolute_error, r2_score
from typing import List, Tuple, Any, Union, Optional, Dict
import warnings
from IPython.utils import io
import logging
import traceback

#mlflow.sklearn.autolog(log_models=True)
logging.basicConfig(level=logging.WARN)

logger = logging.getLogger(__name__)
mlflow.set_tracking_uri(os.environ['MLFLOW_BASE_URL'])
client = mlflow.tracking.MlflowClient(os.environ['MLFLOW_BASE_URL'])

warnings.filterwarnings("ignore")
# * 'lr' - Linear Regression  : done                 
# 'lasso' - Lasso Regression   :  done           
# 'ridge' - Ridge Regression   : done             
# * 'en' - Elastic Net   : done                
# * 'lar' - Least Angle Regression                  
# * 'llar' - Lasso Least Angle Regression                   
# * 'omp' - Orthogonal Matching Pursuit                     
# * 'br' - Bayesian Ridge                   
# * 'ard' - Automatic Relevance Determination                  
# * 'par' - Passive Aggressive Regressor                    
# * 'ransac' - Random Sample Consensus       
# * 'tr' - TheilSen Regressor                   
# * 'huber' - Huber Regressor                               
# * 'kr' - Kernel Ridge                                     
# * 'svm' - Support Vector Regression : done                          
# * 'knn' - K Neighbors Regressor    :done                       
# * 'dt' - Decision Tree Regressor   :done                                
# * 'rf' - Random Forest Regressor : done                                 
# * 'et' - Extra Trees Regressor    :done                        
# * 'ada' - AdaBoost Regressor     : done                         
# * 'gbr' - Gradient Boosting Regressor   :done                            
# * 'mlp' - MLP Regressor
# * 'xgboost' - Extreme Gradient Boosting : done                   
# * 'lightgbm' - Light Gradient Boosting Machine  :   done               
# * 'catboost' - CatBoost Regressor  : done

def input_validation(X_train:pd.DataFrame,
                     X_test:pd.DataFrame,
                     y_train:pd.Series,
                     y_test:pd.Series,
                     ):
        
    if not (isinstance(X_train, pd.DataFrame) or isinstance(X_train,np.ndarray)):
        raise TypeError(f"X_train passed must be of type pandas.DataFrame or numpy.ndarray")
    if X_train.shape[0] == 0:
        raise ValueError(f"X_train passed must be a positive dataframe")
    if not (isinstance(X_test, pd.DataFrame) or isinstance(X_train,np.ndarray)):
        raise TypeError(f"X_test passed must be of type pandas.DataFrame or numpy.ndarray")
    if X_test.shape[0] == 0:
        raise ValueError(f"X_test passed must be a positive dataframe")
    if not (isinstance(y_train,pd.Series) or isinstance(y_train,pd.DataFrame)or isinstance(y_train,np.ndarray)):
        raise TypeError(f"y_train passed must be of type pandas.Series,pandas.DataFrame,numpy.ndarray") 
    if y_train.shape[0] == 0:
        raise ValueError(f"y_train passed must be a positive Series")
    if not (isinstance(y_test,pd.Series) or isinstance(y_test,pd.DataFrame)or isinstance(y_test,np.ndarray)):
        raise TypeError(f"y_test passed must be of type pandas.Series,pandas.DataFrame,numpy.ndarray")
    if y_test.shape[0] == 0:
        raise ValueError(f"y_test passed must be a positive Series")
    if X_train.shape[0]!=len(y_train):
        raise ValueError(f"Length of Train set and its respective Target variable should be of equal length")
    if X_test.shape[0]!=len(y_test):
        raise ValueError(f"Length of Test set and its respective Target variable should be of equal length")
#     if not isinstance(experiment_name,str):
#         raise ValueError(f"experiment name should be string")     

class Regressor(MLClient):
    
    def __init__(self, 
                 X_train:pd.DataFrame,
                 X_test:pd.DataFrame,
                 y_train:pd.Series,
                 y_test:pd.Series,
                 experiment_name:str):
        '''
        Parameters
        ----------
        X_train:pandas.DataFrame
            Shape (n_samples, n_features), where n_samples is the number of samples and 
            n_features is the number of features for train dataset.
        X_test:pandas.DataFrame
            Shape (n_samples, n_features), where n_samples is the number of samples and 
            n_features is the number of features for test dataset
        y_train:pandas.Series
           Shape (n_samples,1),where n_samples is the number of samples for train dataset
        y_test:pandas.Series
           Shape (n_samples,1),where n_samples is the number of samples for test dataset
        experiment_name:str
           Name of the experiment for logging.
           
        '''
        
        logger.info(f"Validation Start")
        input_validation(X_train,X_test, y_train,y_test)
        logger.info(f"Validation End")
        
        super().__init__(experiment_name)
#         self.experiment_name = experiment_name
#         self.experiment_details = get_exp(self.experiment_name)
#         self.experiment_id = self.experiment_details.loc['experiment_id'].values[0]
        
        self.X_train = X_train
        self.X_test = X_test
        self.y_train = y_train
        self.y_test = y_test
        
        logger.info(f"Regression successfully instantiated")

    def __str__(self):
        return "Regression models for experiment: Experiment Name is %s, Experiment ID is %s" % (self.name, self.id)
        
    
    def eval_metrics(self,actual:Union[pd.Series,pd.DataFrame,np.ndarray], pred:Union[pd.Series,pd.DataFrame,np.ndarray])->Tuple:
        '''Calculates various evaluation parameters like mse,rmse,mae rmsle for regression problem
        Parameters
        ----------
        actual: pandas.Series or pandas.DataFrame, or numpy array
             actual testing data
        pred: pandas.Series or pandas.DataFrame, or numpy array
            predicted testing data

        Returns:
        --------

           mse: mean square error
           rmse: root mean square error
           rmsle: root mean square log error
           mae: mean absolute error
           r2: r-2 score
        '''


        logger.info(f"Calculating evaluation paramaters")
        try:
            mse = mean_squared_error(actual, pred)
        except:
            logger.error(f"mean_squared_error calcuation raised an exception:")
            logger.error(traceback.format_exc())
        try:
            rmse = np.sqrt(mse)
        except:
            logger.error(f"root_mean_squared_error calcuation raised an exception:")
            logger.error(traceback.format_exc())
        try:
            rmsle = np.log(rmse)
        except:
            logger.error(f"root_mean_squared_log_error calcuation raised an exception:")
            logger.error(traceback.format_exc())
        try:
            mae = mean_absolute_error(actual, pred)
        except:
            logger.error(f"mean_absolute_error calcuation raised an exception:")
            logger.error(traceback.format_exc())
        try:
            r2 = r2_score(actual, pred)
        except:
            logger.error(f"r2 score calcuation raised an exception:")
            logger.error(traceback.format_exc())
 
        return mse,rmse,rmsle,mae,r2 



    def model_fit_log(self, model:Any,
                     run_name:str,)->None:
        '''
        Performs the model fitting and model prediction task. It calls the evaluation metrics function and logs the required metrics for the model.
        '''
        logger.info(f"Training the model")
        try:
            model.fit(self.X_train,self.y_train)
        except:
            logger.error(f"Fitting the model {run_name} raised an exception:")
            logger.error(traceback.format_exc())

        logger.info(f"Performing the prediction")

        try:
            y_pred=model.predict(self.X_test)
        except:
            logger.error(f" Prediction process raised an exception:")
            logger.error(traceback.format_exc())

        (mse,rmse,rmsle,mae, r2) = self.eval_metrics(self.y_test,y_pred)

        metrics = {"MSE":mse,"RMSE":rmse,"RMSLE":rmsle,"MAE":mae,"R2":r2}

        logger.info(f"logged params: {metrics.keys()}")
        try:
            mlflow.log_metrics(metrics)
        except:
            logger.warning("Couldn't log metrics. Exception:")
            logger.warning(traceback.format_exc())
        logger.info(f"logging model {run_name}")
        try:
            mlflow.sklearn.log_model(model,run_name)
        except:
            logger.warning("Couldn't log model. Exception:")
            logger.warning(traceback.format_exc())    

### Linear Regression####        
    def LinearRegression(self,**kwargs
                         )->None:
        ''' This module performs experiments for Linear Regression model

        **kwargs: 
           List of parameters that will be passed to define the model.
           Refer: https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LinearRegression.html 
           for list of parameters that can be passed to model

         Returns
        -------
                 None
        '''
        logger.info("-----------Linear Regression----------------") 

        from sklearn.linear_model import LinearRegression    

        run_name = self.name+"_"+self.id+"_"+"linear_regression"
        try:
            with mlflow.start_run(run_name=run_name):
                model =LinearRegression(**kwargs)
                self.model_fit_log(model,run_name)
        except:
            logger.warning("Couldn't perform runs on model. Exception:")
            logger.warning(traceback.format_exc())

    def RidgeRegression(self,**kwargs)->None:
        ''' This module performs experiments for Ridge Regression model

        Parameters
        ----------

        **kwargs: 
           List of parameters that will be passed to define the model.
           Refer: https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.Ridge.html 
           for list of parameters that can be passed to model

         Returns
        -------
                 None
        '''
        logger.info("-----------Ridge Regression----------------")
  

        from sklearn.linear_model import Ridge    

        run_name = self.name+"_"+self.id+"_"+"Ridge_regression"
        try:
            with mlflow.start_run(run_name=run_name):
                model =Ridge(**kwargs)
                self.model_fit_log(model,run_name)
        except:
            logger.warning("Couldn't perform runs on model. Exception:")
            logger.warning(traceback.format_exc())   
        
    def LassoRegression(self,**kwargs)->None:
        ''' This module performs experiments for Ridge Regression model

        Parameters
        ----------

        **kwargs: 
           List of parameters that will be passed to define the model.
           Refer: https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.Lasso.html 
           for list of parameters that can be passed to model

         Returns
        -------
                 None
        '''
        logger.info("-----------Lasso Regression----------------")

        from sklearn.linear_model import Lasso    

        run_name = self.name+"_"+self.id+"_"+"Lasso_regression"
        try:
            with mlflow.start_run(run_name=run_name):
                model =Lasso(**kwargs)
                self.model_fit_log(model,run_name)
        except:
            logger.warning("Couldn't perform runs on model. Exception:")
            logger.warning(traceback.format_exc()) 
        
        
    def ElasticNet(self,**kwargs)->None:
        ''' This module performs experiments for Ridge Regression model

        Parameters
        ----------

        **kwargs: 
           List of parameters that will be passed to define the model.
           Refer: https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.ElasticNet.html 
           for list of parameters that can be passed to model

         Returns
        -------
                 None
        '''
        logger.info("-----------Elasticnet Regression----------------")
  
        from sklearn.linear_model import ElasticNet    

        run_name = self.name+"_"+self.id+"_"+"ElasticNet_regression"
        try:
            with mlflow.start_run(run_name=run_name):
                model =ElasticNet(**kwargs)
                self.model_fit_log(model,run_name)
        except:
            logger.warning("Couldn't perform runs on model. Exception:")
            logger.warning(traceback.format_exc())        

    def SupportVectorRegressor(self,**kwargs)->None:

        ''' This module performs experiments for Support Vector Regression model

        Parameters
        ----------
        
        **kwargs: 
           List of parameters that will be passed to define the model.
           Refer: https://scikit-learn.org/stable/modules/svm.html 
           for list of parameters that can be passed to model

         Returns
        -------
                 None
        '''
        logger.info("-----------Support Vector Regression----------------")

        from sklearn.svm import SVR
        
        
        run_name = self.name+"_"+self.id+"_"+"SupportVector_regression"
        try:
            with mlflow.start_run(run_name=run_name):
                model =SVR(**kwargs)
                self.model_fit_log(model,run_name)
        except:
            logger.warning("Couldn't perform runs on model. Exception:")
            logger.warning(traceback.format_exc())

    def KNNRegressor(self,**kwargs)->None:
        ''' This module performs experiments for KNN Regression model

        Parameters
        ----------

        **kwargs: 
           List of parameters that will be passed to define the model.
           Refer: https://scikit-learn.org/stable/modules/generated/sklearn.neighbors.KNeighborsRegressor.html 
           for list of parameters that can be passed to model

         Returns
        -------
                 None
        '''
        logger.info("-----------KNN Regression----------------")
        
           
            

        from sklearn.neighbors import KNeighborsRegressor    

        run_name = self.name+"_"+self.id+"_"+"KNN_regression"
        try:
            with mlflow.start_run(run_name=run_name):
                model =KNeighborsRegressor(**kwargs)
                self.model_fit_log(model,run_name)
        except:
            logger.warning("Couldn't perform runs on model. Exception:")
            logger.warning(traceback.format_exc())   

    def RandomForestRegressor(self,**kwargs)->None:
        ''' This module performs experiments for Random Forest Regression model

        Parameters
        ----------
      
        **kwargs: 
           List of parameters that will be passed to define the model.
           Refer: https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.RandomForestRegressor.html 
           for list of parameters that can be passed to model

         Returns
        -------
                 None
        '''
        logger.info("-----------Random Forest Regression----------------")
        
           
            

        from sklearn.ensemble import RandomForestRegressor    

        

        
        run_name = self.name+"_"+self.id+"_"+"RandomForest_regression"
        try:
            with mlflow.start_run(run_name=run_name):
                model =RandomForestRegressor(**kwargs)
                self.model_fit_log(model,run_name)
        except:
            logger.warning("Couldn't perform runs on model. Exception:")
            logger.warning(traceback.format_exc()) 

    def XGBRegressor(self,**kwargs)->None:
        ''' This module performs experiments for XGBoost Regression model

        Parameters
        ----------

        **kwargs: 
           List of parameters that will be passed to define the model.
           Refer: https://xgboost.readthedocs.io/en/latest/python/python_api.html#module-xgboost.sklearn 
           for list of parameters that can be passed to model

         Returns
        -------
                 None
        '''
        logger.info("-----------XGBoost Regression----------------")
        
        from xgboost import XGBRegressor     

        run_name = self.name+"_"+self.id+"_"+"XGBoost_regression"
        try:
            with mlflow.start_run(run_name=run_name):
                model =XGBRegressor(**kwargs)
                self.model_fit_log(model,run_name)
        except:
            logger.warning("Couldn't perform runs on model. Exception:")
            logger.warning(traceback.format_exc())    

    def CatBoostRegressor(self,**kwargs)->None:
        ''' This module performs experiments for XGBoost Regression model

        Parameters
        ----------

        **kwargs: 
           List of parameters that will be passed to define the model.
           Refer: https://catboost.ai/en/docs/concepts/python-reference_catboostregressor 
           for list of parameters that can be passed to model

         Returns
        -------
                 None
        '''
        logger.info("-----------Catboost Regression----------------")
        
        from catboost import CatBoostRegressor     
        
        run_name = self.name+"_"+self.id+"_"+"CatBoost_regression"
        try:
            with mlflow.start_run(run_name=run_name):
                model =CatBoostRegressor(**kwargs)
                self.model_fit_log(model,run_name)
        except:
            logger.warning("Couldn't perform runs on model. Exception:")
            logger.warning(traceback.format_exc())        

    def LGBMRegressor(self,**kwargs)->None:
        ''' This module performs experiments for LGBM Regression model

        Parameters
        ----------

        **kwargs: 
           List of parameters that will be passed to define the model.
           Refer: https://lightgbm.readthedocs.io/en/latest/pythonapi/lightgbm.LGBMRegressor.html 
           for list of parameters that can be passed to model

         Returns
        -------
                 None
        '''
        logger.info("-----------LightGBM Regression----------------")
        
        from lightgbm import LGBMRegressor     

        run_name = self.name+"_"+self.id+"_"+"LGBM_regression"
        try:
            with mlflow.start_run(run_name=run_name):
                model =LGBMRegressor(**kwargs)
                self.model_fit_log(model,run_name)
        except:
            logger.warning("Couldn't perform runs on model. Exception:")
            logger.warning(traceback.format_exc())        


    def GradientBoostingRegressor(self,**kwargs)->None:
        ''' This module performs experiments for Gradient Boosting Regression model

        Parameters
        ----------

        **kwargs: 
           List of parameters that will be passed to define the model.
           Refer: https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.GradientBoostingRegressor.html 
           for list of parameters that can be passed to model

         Returns
        -------
                 None
        '''
        logger.info("-----------GradientBoost Regression----------------")
        
        from sklearn.ensemble import GradientBoostingRegressor     

        run_name = self.name+"_"+self.id+"_"+"GB_regression"
        try:
            with mlflow.start_run(run_name=run_name):
                model =GradientBoostingRegressor(**kwargs)
                self.model_fit_log(model,run_name)
        except:
            logger.warning("Couldn't perform runs on model. Exception:")
            logger.warning(traceback.format_exc())  

    def AdaBoostRegressor( self,**kwargs)->None:
        ''' This module performs experiments for AdaBoost Regression model

        Parameters
        ----------

        **kwargs: 
           List of parameters that will be passed to define the model.
           Refer: https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.AdaBoostRegressor.html 
           for list of parameters that can be passed to model

         Returns
        -------
                 None
        '''
        logger.info("-----------AdaBoost Regression----------------")
       
        from sklearn.ensemble import AdaBoostRegressor     

        run_name = self.name+"_"+self.id+"_"+"Adaboost_regression"
        try:
            with mlflow.start_run(run_name=run_name):
                model =AdaBoostRegressor(**kwargs)
                self.model_fit_log(model,run_name)
        except:
            logger.warning("Couldn't perform runs on model. Exception:")
            logger.warning(traceback.format_exc())    

    def DecisionTreeRegressor(self,**kwargs)->None:
        ''' This module performs experiments for DecisionTree Regression model

        Parameters
        ----------

        **kwargs: 
           List of parameters that will be passed to define the model.
           Refer: https://scikit-learn.org/stable/modules/generated/sklearn.tree.DecisionTreeRegressor.html 
           for list of parameters that can be passed to model

         Returns
        -------
                 None
        '''
        logger.info("-----------DecisionTree Regression----------------")
        
        from sklearn.tree import DecisionTreeRegressor     

        run_name = self.name+"_"+self.id+"_"+"DecisionTree_regression"
        try:
            with mlflow.start_run(run_name=run_name):
                model =DecisionTreeRegressor(**kwargs)
                self.model_fit_log(model,run_name)
        except:
            logger.warning("Couldn't perform runs on model. Exception:")
            logger.warning(traceback.format_exc())  


    def ExtraTreeRegressor(self,**kwargs)->None:
        ''' This module performs experiments for ExtraTree Regression model

        Parameters
        ----------
        
        **kwargs: 
           List of parameters that will be passed to define the model.
           Refer: https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.ExtraTreesRegressor.html 
           for list of parameters that can be passed to model

         Returns
        -------
                 None
        '''
        logger.info("-----------ExtraTree Regression----------------")
        
          
        from sklearn.ensemble import ExtraTreesRegressor     

        run_name = self.name+"_"+self.id+"_"+"ExtraTree_regression"
        try:
            with mlflow.start_run(run_name=run_name):
                model =ExtraTreesRegressor(**kwargs)
                self.model_fit_log(model,run_name)
        except:
            logger.warning("Couldn't perform runs on model. Exception:")
            logger.warning(traceback.format_exc())        