
import os
from typing import List
import mlflow
import pandas as pd
import joblib

mlflow.set_tracking_uri(os.environ['MLFLOW_BASE_URL'])
client = mlflow.tracking.MlflowClient(os.environ['MLFLOW_BASE_URL'])

def set_exp(exp_name:str) -> None:
    '''
    Set given experiment as active experiment. If experiment does not exist, create an experiment with provided name.
    exp_name – Case sensitive name of an experiment to be activated.
    '''
    return mlflow.set_experiment(exp_name)
    
def get_exp(exp_name:str) :
    '''
    Retrieve an experiment by experiment name from the backend store.
    exp_name – The case senstive experiment name.
    '''
    result ={}
    mlflow.set_experiment(exp_name)
    exp_details = mlflow.get_experiment_by_name(exp_name)
    result['experiment_name'] = exp_details.name
    result['location']=exp_details.artifact_location
    result['experiment_id'] = exp_details.experiment_id
    result['experiment_stage'] = exp_details.lifecycle_stage
    result['tags'] = exp_details.tags
    
    return pd.DataFrame.from_dict(result,orient='index',columns=['parameters'])
    
#def get_exp_id(exp_name:str) :
#    '''
#    Retrieve an experiment id by experiment name from the backend store.
#    exp_name – The case senstive experiment name.
#    '''
#    mlflow.set_experiment(exp_name)
#    exp_details = mlflow.get_experiment_by_name(exp_name)
#    exp_id = exp_details.experiment_id
#    return exp_id
    
    
def staging(
            run_id:str,
            exp_name:str,
            model_name:str,
            stage:str
           ) -> None:
    '''
    register the model in model registry and transit it to given stage.
    Staging, Production or Archived.
    '''
    
    result = mlflow.register_model(
    "runs:/" + run_id,
    model_name
    )
    registered_model = pd.DataFrame(result, columns =['Attribute','Value']).set_index('Attribute')
    staging = client.transition_model_version_stage(
    name=registered_model.loc['name'][0],
    version=registered_model.loc['version'][0],
    stage=stage,
    )
    staged_model = pd.DataFrame(staging, columns =['Attribute','Value']).set_index('Attribute')
    print('Model information :')
    print(staged_model)
    return staged_model
    
def change_stage(
                 model_name:str,
                 ver_list:List[str],
                 stage:str
                ) -> None:
    '''
    change stage of model.(Staging, Production or Archived.)
    '''
    
    for i in ver_list:
        staging = client.transition_model_version_stage(
        name=model_name,
        version=i,
        stage=stage,
        )
        print('{0} verison {1} archived'.format(model_name,i))

def model_versions(
                   model_name:str
                   )-> None:
    '''
    return model versions if match with filter string.
    '''
    filter_string = "name='{}'".format(model_name)
    results = client.search_model_versions(filter_string)
    for res in results:
        print("name={}; run_id={}; version={}".format(res.name, res.run_id, res.version))
        
def log_data(
             prun_id:str,
             data_obj,
             location:str) ->None:
    '''
    store joblib objects in model artifact.
    '''
    joblib.dump(data_obj,location)
    with mlflow.start_run(run_id = prun_id):
        mlflow.log_artifact(location,artifact_path='model/'+location+'/')
    print('data saved successfully....')
        
def search_runs(exp_id:str):
    '''
    Search runs and return dataframe of runs.
    '''
    df = mlflow.search_runs(experiment_ids=exp_id)
    df.columns = df.columns.str.replace('tags.mlflow.runName','run_name')
    exclude_cols = ['tags.mlflow.source.type','tags.mlflow.user', 'tags.mlflow.source.name']
    #column_names =  [x for x not in exclude_cols]
    df = df[df.columns.difference(exclude_cols)]
    return  df


def previous_version(
                    model_name:str
                    ):
    '''
    return list of all previous versions.
    '''
    
    all_versions = []
    filter_string = "name='{}'".format(model_name)
    for x in client.search_model_versions(filter_string):
        all_versions.append(int(dict(x)['version']))
    all_versions.sort()
    previous_versions = list(map(lambda x : str(x),all_versions))[:-1]
    return previous_versions

def delete_model_version(
                        model_name:str,
                        ver_list:List[str]
                        ):
    '''
    delete model versions.
    '''
    for version in ver_list:
        client.delete_model_version(name=model_name, version=version)
    
    return None

def delete_reg_model(
                    model_name:str
                    ):
    '''
    delete registered model with all its version.
    '''
    client.delete_registered_model(name=model_name)
    
    return None

def delete_exp(
                    experiment_id:str
                    ):
    '''
    delete registered model with all its version.
    '''
    client.delete_experiment(experiment_id)
    
    return None

def delete_run_by_id(run_id:str):
    client.delete_run(run_id)
    return None

def load_model(location):#(model_name,stage):
    return mlflow.sklearn.load_model(location)

    
    

